import logging
import sys
import traceback

from django import db
from dramatiq.middleware import Middleware

LOGGER = logging.getLogger("django_dramatiq.AdminMiddleware")


class AdminMiddleware(Middleware):
    """This middleware keeps track of task executions.
    """

    def after_enqueue(self, broker, message, delay):
        from .models import Task

        LOGGER.debug("Creating Task from message %r.", message.message_id)
        status = Task.STATUS_ENQUEUED
        if delay:
            status = Task.STATUS_DELAYED

        Task.tasks.create_or_update_from_message(
            message,
            status=status,
            actor_name=message.actor_name,
            queue_name=message.queue_name,
        )

    def before_process_message(self, broker, message):
        from .models import Task

        LOGGER.debug("Updating Task from message %r.", message.message_id)
        Task.tasks.create_or_update_from_message(
            message,
            status=Task.STATUS_RUNNING,
            actor_name=message.actor_name,
            queue_name=message.queue_name,
        )

    def after_skip_message(self, broker, message):
        from .models import Task

        self.after_process_message(broker, message, status=Task.STATUS_SKIPPED)

    def after_process_message(self, broker, message, *, result=None, exception=None, status=None):
        from .models import Task

        if exception is not None:
            status = Task.STATUS_FAILED
            exc_type, exc_value, exc_traceback = sys.exc_info()
            formatted_exception = traceback.format_exception(
                exception,
                exc_value,
                exc_traceback,
                limit=30,
            )
            message.options["traceback"] = "".join(formatted_exception)
        elif status is None:
            status = Task.STATUS_DONE

        LOGGER.debug("Updating Task from message %r.", message.message_id)
        Task.tasks.create_or_update_from_message(
            message,
            status=status,
            actor_name=message.actor_name,
            queue_name=message.queue_name,
        )


class DbConnectionsMiddleware(Middleware):
    """This middleware cleans up db connections on worker shutdown.
    """

    def _close_old_connections(self, *args, **kwargs):
        db.close_old_connections()

    before_process_message = _close_old_connections
    after_process_message = _close_old_connections

    def _close_connections(self, *args, **kwargs):
        db.connections.close_all()

    before_consumer_thread_shutdown = _close_connections
    before_worker_thread_shutdown = _close_connections
    before_worker_shutdown = _close_connections



import uuid
from datetime import timedelta

import pytest
from django.utils.timezone import now

from django_dramatiq.models import Task
from django_dramatiq.tasks import delete_old_tasks


def test_can_delete_old_tasks(db):
    # Given a Task that was created more than a day ago
    task = Task(id=uuid.uuid4(), message_data=b"")
    task.save()

    # we have to set the created_at time after the initial save in
    # order to avoid it getting overwritten.
    task.created_at = now() - timedelta(days=2)
    task.save()

    # When I call the delete_old_tasks task
    delete_old_tasks()

    # Then my task should be deleted
    with pytest.raises(Task.DoesNotExist):
        task.refresh_from_db()



import sys
from threading import Event

import dramatiq
import pytest
from dramatiq import Middleware
from dramatiq.middleware import SkipMessage

from django_dramatiq.models import Task


def test_admin_middleware_keeps_track_of_tasks(transactional_db, broker, worker):
    # Given an actor
    evt = Event()

    @dramatiq.actor
    def do_work():
        evt.set()

    # When I send it a delayed message
    do_work.send_with_options(delay=250)

    # Then a Task should be stored to the database
    task = Task.tasks.get()
    assert task
    assert task.status == Task.STATUS_DELAYED

    # When I join on the broker
    evt.wait()
    broker.join(do_work.queue_name, fail_fast=True)
    worker.join()

    # And reload the task
    task.refresh_from_db()

    # Then the Task's status should be updated
    assert task.status == Task.STATUS_DONE


@pytest.mark.skipif(sys.version_info < (3, 11), reason="flaky")
def test_admin_middleware_keeps_track_of_failed_tasks(transactional_db, broker, worker):
    # Given an actor that always fails
    @dramatiq.actor(max_retries=0)
    def do_work():
        raise RuntimeError("failed")

    # When I send it a message
    do_work.send()

    # And I join on the broker
    broker.join(do_work.queue_name)
    worker.join()

    # Then a failed Task should be stored to the database
    task = Task.tasks.get()
    assert task
    assert task.status == Task.STATUS_FAILED
    assert "RuntimeError" in task.message.options["traceback"]


def test_admin_middleware_keeps_track_of_skipped_tasks(transactional_db, broker, worker):
    # Given an actor that does nothing
    @dramatiq.actor(max_retries=0)
    def do_work():
        pass

    # And a middleware that skips all messages
    class Skipper(Middleware):
        def before_process_message(self, broker, message):
            raise SkipMessage()

    # When I enable the middleware
    broker.add_middleware(Skipper())

    # And I send the actor a message
    do_work.send()

    # And I join on the broker
    broker.join(do_work.queue_name)
    worker.join()

    # Then a skipped Task should be stored to the database
    task = Task.tasks.get()
    assert task
    assert task.status == Task.STATUS_SKIPPED



import dramatiq

from django_dramatiq.models import Task
from django_dramatiq.test import DramatiqTestCase


class TestDramatiqTestCase(DramatiqTestCase):

    def test_worker_consumes_tasks(self):
        # Given an actor defined in the test method
        @dramatiq.actor(max_retries=0)
        def do_work():
            self.assertEqual(1, Task.tasks.count())

        # When I send it a message
        do_work.send()

        # And I join on the broker
        self.broker.join(do_work.queue_name)
        self.worker.join()

        # Then a finished Task should be stored to the database
        task = Task.tasks.get()
        self.assertIsNotNone(task)
        self.assertEqual(task.status, Task.STATUS_DONE)



import os

from setuptools import setup


def rel(*xs):
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), *xs)


with open(rel("django_dramatiq", "__init__.py"), "r") as f:
    version_marker = "__version__ = "
    for line in f:
        if line.startswith(version_marker):
            _, version = line.split(version_marker)
            version = version.strip().strip('"')
            break
    else:
        raise RuntimeError("Version marker not found.")


setup(
    name="django_dramatiq",
    version=version,
    description="A Django app for Dramatiq.",
    long_description="Visit https://github.com/Bogdanp/django_dramatiq for more information.",
    packages=[
        "django_dramatiq",
        "django_dramatiq.management",
        "django_dramatiq.management.commands",
        "django_dramatiq.migrations",
    ],
    install_requires=[
        "django>=3.2",
        "dramatiq>=1.11",
    ],
    classifiers=[
        "Environment :: Web Environment",
        "Operating System :: OS Independent",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.1",
        "Framework :: Django :: 4.2",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    extras_require={
        "dev": [
            "bumpversion",
            "flake8",
            "flake8-quotes",
            "isort",
            "pytest",
            "pytest-cov",
            "pytest-django",
            "twine",
        ]
    },
    python_requires=">=3.8",
    include_package_data=True,
)



import os
import sys
from io import StringIO
from unittest.mock import patch

from django.core.management import call_command

from django_dramatiq.management.commands import rundramatiq


def test_rundramatiq_command_autodiscovers_modules():
    assert rundramatiq.Command().discover_tasks_modules() == [
        "django_dramatiq.setup",
        "django_dramatiq.tasks",
        "tests.testapp1.tasks",
        "tests.testapp2.tasks",
        "tests.testapp3.tasks.other_tasks",
        "tests.testapp3.tasks.tasks",
        "tests.testapp3.tasks.utils",
        "tests.testapp3.tasks.utils.not_a_task",
    ]


@patch("os.execvp")
def test_rundramatiq_can_run_dramatiq(execvp_mock):
    # Given an output buffer
    buff = StringIO()

    # When I call the rundramatiq command
    call_command("rundramatiq", stdout=buff)

    # Then stdout should contain a message about discovered task modules
    assert "Discovered tasks module: 'tests.testapp1.tasks'" in buff.getvalue()
    assert "Discovered tasks module: 'tests.testapp2.tasks'" in buff.getvalue()
    assert "Discovered tasks module: 'tests.testapp3.tasks.tasks'" in buff.getvalue()
    assert "Discovered tasks module: 'tests.testapp3.tasks.other_tasks'" in buff.getvalue()

    # And execvp should be called with the appropriate arguments
    cores = str(rundramatiq.CPU_COUNT)
    expected_exec_name = "dramatiq"
    expected_exec_path = os.path.join(
        os.path.dirname(sys.executable),
        expected_exec_name,
    )

    execvp_mock.assert_called_once_with(expected_exec_path, [
        expected_exec_name, "--path", ".", "--processes", cores, "--threads", cores,
        "--worker-shutdown-timeout", "600000",
        "django_dramatiq.setup",
        "django_dramatiq.tasks",
        "tests.testapp1.tasks",
        "tests.testapp2.tasks",
        "tests.testapp3.tasks.other_tasks",
        "tests.testapp3.tasks.tasks",
        "tests.testapp3.tasks.utils",
        "tests.testapp3.tasks.utils.not_a_task",
    ])


@patch("os.execvp")
def test_rundramatiq_can_run_dramatiq_reload(execvp_mock):
    # Given an output buffer
    buff = StringIO()

    # When I call the rundramatiq command with --reload-use-polling
    call_command("rundramatiq", "--reload", stdout=buff)

    # Then execvp should be called with the appropriate arguments
    cores = str(rundramatiq.CPU_COUNT)
    expected_exec_name = "dramatiq"
    expected_exec_path = os.path.join(
        os.path.dirname(sys.executable),
        expected_exec_name,
    )

    execvp_mock.assert_called_once_with(expected_exec_path, [
        expected_exec_name, "--path", ".", "--processes", cores, "--threads", cores,
        "--worker-shutdown-timeout", "600000",
        "--watch", ".",
        "django_dramatiq.setup",
        "django_dramatiq.tasks",
        "tests.testapp1.tasks",
        "tests.testapp2.tasks",
        "tests.testapp3.tasks.other_tasks",
        "tests.testapp3.tasks.tasks",
        "tests.testapp3.tasks.utils",
        "tests.testapp3.tasks.utils.not_a_task",
    ])


@patch("os.execvp")
def test_rundramatiq_can_run_dramatiq_with_polling(execvp_mock):
    # Given an output buffer
    buff = StringIO()

    # When I call the rundramatiq command with --reload-use-polling
    call_command("rundramatiq", "--reload", "--reload-use-polling", stdout=buff)

    # Then execvp should be called with the appropriate arguments
    cores = str(rundramatiq.CPU_COUNT)
    expected_exec_name = "dramatiq"
    expected_exec_path = os.path.join(
        os.path.dirname(sys.executable),
        expected_exec_name,
    )

    execvp_mock.assert_called_once_with(expected_exec_path, [
        expected_exec_name, "--path", ".", "--processes", cores, "--threads", cores,
        "--worker-shutdown-timeout", "600000",
        "--watch", ".",
        "--watch-use-polling",
        "django_dramatiq.setup",
        "django_dramatiq.tasks",
        "tests.testapp1.tasks",
        "tests.testapp2.tasks",
        "tests.testapp3.tasks.other_tasks",
        "tests.testapp3.tasks.tasks",
        "tests.testapp3.tasks.utils",
        "tests.testapp3.tasks.utils.not_a_task",
    ])


@patch("os.execvp")
def test_rundramatiq_can_run_dramatiq_with_only_some_queues(execvp_mock):
    # Given an output buffer
    buff = StringIO()

    # When I call the rundramatiq command with --queues
    call_command("rundramatiq", "--queues", "A B C", stdout=buff)

    # Then execvp should be called with the appropriate arguments
    cores = str(rundramatiq.CPU_COUNT)
    expected_exec_name = "dramatiq"
    expected_exec_path = os.path.join(
        os.path.dirname(sys.executable),
        expected_exec_name,
    )

    execvp_mock.assert_called_once_with(expected_exec_path, [
        expected_exec_name, "--path", ".", "--processes", cores, "--threads", cores,
        "--worker-shutdown-timeout", "600000",
        "django_dramatiq.setup",
        "django_dramatiq.tasks",
        "tests.testapp1.tasks",
        "tests.testapp2.tasks",
        "tests.testapp3.tasks.other_tasks",
        "tests.testapp3.tasks.tasks",
        "tests.testapp3.tasks.utils",
        "tests.testapp3.tasks.utils.not_a_task",
        "--queues", "A B C"
    ])


@patch("os.execvp")
def test_rundramatiq_can_run_dramatiq_with_specified_pid_file(execvp_mock):
    # Given an output buffer
    buff = StringIO()

    # When I call the rundramatiq command with --pid-file
    call_command("rundramatiq", "--pid-file", "drama.pid", stdout=buff)

    # Then execvp should be called with the appropriate arguments
    cores = str(rundramatiq.CPU_COUNT)
    expected_exec_name = "dramatiq"
    expected_exec_path = os.path.join(
        os.path.dirname(sys.executable),
        expected_exec_name,
    )

    execvp_mock.assert_called_once_with(expected_exec_path, [
        expected_exec_name, "--path", ".", "--processes", cores, "--threads", cores,
        "--worker-shutdown-timeout", "600000",
        "django_dramatiq.setup",
        "django_dramatiq.tasks",
        "tests.testapp1.tasks",
        "tests.testapp2.tasks",
        "tests.testapp3.tasks.other_tasks",
        "tests.testapp3.tasks.tasks",
        "tests.testapp3.tasks.utils",
        "tests.testapp3.tasks.utils.not_a_task",
        "--pid-file", "drama.pid"
    ])


@patch("os.execvp")
def test_rundramatiq_can_run_dramatiq_with_specified_log_file(execvp_mock):
    # Given an output buffer
    buff = StringIO()

    # When I call the rundramatiq command with --log-file
    call_command("rundramatiq", "--log-file", "drama.log", stdout=buff)

    # Then execvp should be called with the appropriate arguments
    cores = str(rundramatiq.CPU_COUNT)
    expected_exec_name = "dramatiq"
    expected_exec_path = os.path.join(
        os.path.dirname(sys.executable),
        expected_exec_name,
    )

    execvp_mock.assert_called_once_with(expected_exec_path, [
        expected_exec_name, "--path", ".", "--processes", cores, "--threads", cores,
        "--worker-shutdown-timeout", "600000",
        "django_dramatiq.setup",
        "django_dramatiq.tasks",
        "tests.testapp1.tasks",
        "tests.testapp2.tasks",
        "tests.testapp3.tasks.other_tasks",
        "tests.testapp3.tasks.tasks",
        "tests.testapp3.tasks.utils",
        "tests.testapp3.tasks.utils.not_a_task",
        "--log-file", "drama.log"
    ])


@patch("os.execvp")
def test_rundramatiq_can_ignore_modules(execvp_mock, settings):
    # Given an output buffer
    buff = StringIO()

    # And 'tests.testapp2.tasks' in DRAMATIQ_IGNORED_MODULES
    # And 'tests.testapp3.tasks.other_tasks' in DRAMATIQ_IGNORED_MODULES
    settings.DRAMATIQ_IGNORED_MODULES = (
        "tests.testapp2.tasks",
        "tests.testapp3.tasks.other_tasks",
        "tests.testapp3.tasks.utils.*",
    )

    # When I call the rundramatiq command
    call_command("rundramatiq", stdout=buff)

    # Then stdout should contain a message about ignored task modules
    assert "Discovered tasks module: 'tests.testapp1.tasks'" in buff.getvalue()
    assert "Discovered tasks module: 'tests.testapp3.tasks.tasks'" in buff.getvalue()
    assert "Discovered tasks module: 'tests.testapp3.tasks.utils'" in buff.getvalue()
    assert "Ignored tasks module: 'tests.testapp2.tasks'" in buff.getvalue()
    assert "Ignored tasks module: 'tests.testapp3.tasks.other_tasks'" in buff.getvalue()
    assert "Ignored tasks module: 'tests.testapp3.tasks.utils.not_a_task'" in buff.getvalue()

    # And execvp should be called with the appropriate arguments
    cores = str(rundramatiq.CPU_COUNT)
    expected_exec_name = "dramatiq"
    expected_exec_path = os.path.join(
        os.path.dirname(sys.executable),
        expected_exec_name,
    )

    execvp_mock.assert_called_once_with(expected_exec_path, [
        expected_exec_name, "--path", ".", "--processes", cores, "--threads", cores,
        "--worker-shutdown-timeout", "600000",
        "django_dramatiq.setup",
        "django_dramatiq.tasks",
        "tests.testapp1.tasks",
        "tests.testapp3.tasks.tasks",
        "tests.testapp3.tasks.utils",
    ])


@patch("os.execvp")
def test_rundramatiq_can_fork(execvp_mock, settings):
    # Given an output buffer
    buff = StringIO()

    # When I call the rundramatiq command with --log-file
    call_command("rundramatiq", "--fork-function", "a", "--fork-function", "b", stdout=buff)

    # Then execvp should be called with the appropriate arguments
    cores = str(rundramatiq.CPU_COUNT)
    expected_exec_name = "dramatiq"
    expected_exec_path = os.path.join(
        os.path.dirname(sys.executable),
        expected_exec_name,
    )

    execvp_mock.assert_called_once_with(expected_exec_path, [
        expected_exec_name, "--path", ".", "--processes", cores, "--threads", cores,
        "--worker-shutdown-timeout", "600000",
        "--fork-function", "a",
        "--fork-function", "b",
        "django_dramatiq.setup",
        "django_dramatiq.tasks",
        "tests.testapp1.tasks",
        "tests.testapp2.tasks",
        "tests.testapp3.tasks.other_tasks",
        "tests.testapp3.tasks.tasks",
        "tests.testapp3.tasks.utils",
        "tests.testapp3.tasks.utils.not_a_task",
    ])



import uuid
from unittest import mock

from django_dramatiq.models import Task


def test_task_create_or_update_from_message(transactional_db, broker, worker):
    message = mock.Mock()
    message_id = uuid.uuid4()
    message.encode.return_value = b"{}"
    message.message_id = message_id

    Task.tasks.create_or_update_from_message(message)

    # Created it
    t = Task.tasks.get(pk=message.message_id)
    message.encode.assert_called_once_with()
    assert t.message_data == message.encode.return_value
    message.encode.reset_mock()
    message.encode.return_value = b'{"another_one", 12}'
    Task.tasks.create_or_update_from_message(message)

    # Updated it
    t.refresh_from_db()
    message.encode.assert_called_once_with()
    assert Task.tasks.count() == 1
    assert t.message_data == message.encode.return_value
